#!/usr/bin/env python3
"""
Command line tool for MetSim
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import logging
import os
import sys
from collections import OrderedDict
from configparser import SafeConfigParser


def _is_valid_file(parser, arg):
    if not os.path.isfile(arg):
        message = "Configuration file {} does not exist.\n".format(arg)
        parser.print_help()
        parser.exit(status=1, message=message)
    return arg


def parse(args):
    """Parse the command line arguments"""
    parser = argparse.ArgumentParser()
    parser.add_argument('config', type=lambda x: _is_valid_file(parser, x),
                        help='Input configuration file')
    parser.add_argument('-n', '--n-processes', default=1, type=int,
                        help='Parallel mode: number of processes to use')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Increase the verbosity of MetSim')
    parser.add_argument('-tg', '--time_grouper', nargs='?',
                        const='1AS', default=None, type=str,
                        help='Pandas TimeGrouper string (e.g. `1AS`)')
    return parser.parse_args()


def init(opts):
    """Initialize some information based on the options & config"""
    config = SafeConfigParser()
    config.optionxform = str
    config.read(opts.config)
    conf = OrderedDict(config['MetSim'])
    conf['forcing_vars'] = OrderedDict(config['forcing_vars'])
    conf['domain_vars'] = OrderedDict(config['domain_vars'])
    conf['state_vars'] = OrderedDict(config['state_vars'])
    out_dir = conf['out_dir']
    out_state = conf.get('out_state', None)
    if out_state is None:
        out_state = os.path.join(out_dir, 'state.nc')

    method = conf['method']

    # If the forcing variable is a directory, scan it for files
    if os.path.isdir(conf['forcing']):
        forcing_files = [os.path.join(conf['forcing'], fn) for fn in
                         next(os.walk(conf['forcing']))[2]]
    else:
        forcing_files = conf['forcing']

    # We assume there is only one domain file and one state file
    domain_file = conf['domain']
    state_file = conf['state']

    def to_list(s):
        return json.loads(s.replace("'", '"'))

    conf.update({"calendar": conf.get('calendar', 'standard'),
                 "nprocs": opts.n_processes,
                 "method": method,
                 "out_dir": out_dir,
                 "out_state": out_state,
                 "state": state_file,
                 "domain": domain_file,
                 "forcing": forcing_files,
                 "verbose": opts.verbose * logging.INFO})
    conf['out_vars'] = to_list(conf.get('out_vars', '[]'))
    conf['iter_dims'] = to_list(conf.get('iter_dims', '["lat", "lon"]'))
    if opts.time_grouper is not None:
        conf['time_grouper'] = opts.time_grouper
    conf = {k: v for k, v in conf.items() if v != []}
    return conf


def main():
    """Runs MetSim"""
    from metsim.metsim import MetSim
    setup = init(parse(sys.argv[1:]))
    ms = MetSim(setup)
    if ms.params['nprocs'] > 1:
        ms.launch()
    else:
        ms.run()


if __name__ == '__main__':
    main()



"""
Stores default constants
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    # prefer to just use the dictionary stored in netCDF
    from netCDF4 import default_fillvals as FILL_VALUES
except ImportError:
    # but we can use these if we can't import netCDF4
    FILL_VALUES = {'S1': '\x00',
                   'f4': 9.969209968386869e+36,
                   'f8': 9.969209968386869e+36,
                   'i1': -127,
                   'i2': -32767,
                   'i4': -2147483647,
                   'i8': -9223372036854775806,
                   'u1': 255,
                   'u2': 65535,
                   'u4': 4294967295,
                   'u8': 18446744073709551614}

DEG_PER_REV = 360.0       # Number of degrees in full revolution
SEC_PER_RAD = 13750.9871  # seconds per radian of hour angle
RAD_PER_DAY = 0.017214    # radians of Earth orbit per julian day
RAD_PER_DEG = 0.01745329  # radians per degree
MIN_DECL = -0.4092797     # minimum declination (radians)
DAYS_OFF = 11.25          # julian day offset of winter solstice
# Note:  Make sure that 3600 % SW_RAD_DT == 0
SW_RAD_DT = 30.0          # timestep for radiation routine (seconds)

MA = 28.9644e-3         # (kg mol-1) molecular weight of air
R = 8.3143              # (m3 Pa mol-1 K-1) gas law constant
R_DRY = 287             # (J / degC * kg) Gas constant of dry air
G_STD = 9.80665         # (m s-2) standard gravitational accel.
P_STD = 101325.0        # (Pa) standard pressure at 0. m elevation
T_STD = 288.15          # (K) standard temp at 0. m elevation
CP = 1010.0             # (J kg-1 K-1) specific heat of air

KELVIN = 273.15
EPS = 0.62196351
DAYS_PER_YEAR = 365.25
SEC_PER_DAY = 86400
SEC_PER_MIN = 60
MIN_PER_HOUR = 60
HOURS_PER_DAY = 24
MIN_PER_DAY = 1440
SEC_PER_HOUR = 3600
STEFAN_B = 5.669e-8      # (W m^-22 K^-4) Stefan Boltzmann constant

M_PER_KM = 1000.0
MBAR_PER_BAR = 1000.0
MM_PER_CM = 10.0
MAX_PERCENT = 100.0

# parameters for the radiation algorithm
# (dim) stands for dimensionless values
C = 1.5             # (dim) radiation parameter
B0 = 0.031          # (dim) radiation parameter
B1 = 0.201          # (dim) radiation parameter
B2 = 0.185          # (dim) radiation parameter
TBASE = 0.870       # (dim) max inst. trans. 0m nadir dry atm
ABASE = -6.1e-5     # (1/Pa) vapor pressure effect on transmittance
SC_INT = 1.32       # (MJ/m2/day) snow correction intercept
SC_SLOPE = 0.096    # (MJ/m2/day/cm) snow correction slope



import re

import numpy as np
import pandas as pd
import xarray as xr
from netCDF4 import date2num, num2date

DEFAULT_ORIGIN = '0001-01-01'


def date_range(start=None, end=None, periods=None, freq='D', tz=None,
               normalize=False, name=None, closed=None, calendar='standard',
               **kwargs):
    ''' Return a fixed frequency datetime index, with day (calendar) as the
    default frequency

    Parameters
    ----------
    start : string or datetime-like, default None
        Left bound for generating dates
    end : string or datetime-like, default None
        Right bound for generating dates
    periods : integer or None, default None
        If None, must specify start and end
    freq : string or DateOffset, default 'D' (calendar daily)
        Frequency strings can have multiples, e.g. '5H'
    tz : string or None
        Time zone name for returning localized DatetimeIndex, for example
        Asia/Hong_Kong
    normalize : bool, default False
        Normalize start/end dates to midnight before generating date range
    name : str, default None
        Name of the resulting index
    closed : string or None, default None
        Make the interval closed with respect to the given frequency to
        the 'left', 'right', or both sides (None)
    calendar : string
        Describes the calendar used in the time calculations. Default is a the
        standard calendar (with leap years)

    Notes
    -----
    2 of start, end, or periods must be specified
    To learn more about the frequency strings, please see `this link
    <http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases>`__.

    Returns
    -------
    rng : DatetimeIndex
    '''
    if calendar in ['standard', 'gregorian', 'propoleptic_gregorian']:
        return pd.date_range(start=start, end=end, periods=periods,
                             freq=freq, tz=tz, normalize=normalize, name=name,
                             closed=closed, **kwargs)
    else:
        # start and end are give
        if (start is not None) and (end is not None) and (periods is None):

            steps, units = decode_freq(freq)
            start_num, end_num = date2num(
                pd.to_datetime([start, end]).to_pydatetime(),
                units, calendar=calendar)
            # Todo divide by freq
            periods = int((end_num - start_num) / steps) + 1

            times = num2date(
                np.linspace(start_num, end_num, periods,
                            endpoint=True,
                            dtype=np.float128), units, calendar)

            index = pd.DatetimeIndex(xr.conventions.nctime_to_nptime(times))

            return index

        else:
            raise NotImplementedError(
                'Specified arguments are not valid for this calendar')


def decode_freq(freq):
    if len(freq) > 1:
        r = re.compile('([0-9]+)([a-zA-Z]+)')
        step, unit = r.match(freq).groups()
    else:
        step = 1
        unit = freq
    return (int(step), units_from_freq(unit))


def units_from_freq(freq, origin=DEFAULT_ORIGIN):
    if 'H' in freq:
        return 'hours since %s' % origin
    elif 'D' in freq:
        return 'days since %s' % origin
    elif 'T' in freq:
        return 'minutes since %s' % origin
    else:
        raise NotImplementedError(
            'freq {} not supported at this time'.format(freq))



"""
IO Module for MetSim
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

import os
import struct

import numpy as np
import pandas as pd
import xarray as xr

from metsim.datetime import date_range


def read_met_data(params: dict, domain: xr.Dataset) -> xr.Dataset:
    """
    Read input meteorological forcings for MetSim.
    This method supports ascii, binary, netcdf, and
    xarray input pointers.  The input source is derived
    from the key 'forcing' in the params dictionary.
    The format of the data is derived from 'in_format'
    key in the parameter dictionary.
    """
    process_funcs = {
        "netcdf": process_nc,
        "binary": process_vic,
        "ascii": process_vic,
        "data": process_nc
    }
    return process_funcs[params['forcing_fmt']](params, domain)


def read_domain(params: dict) -> xr.Dataset:
    """Load in a domain file"""
    read_funcs = {
        "netcdf": read_netcdf,
        "data": read_data
    }
    return read_funcs[params['domain_fmt']](
        params['domain'], calendar=params['calendar'],
        var_dict=params.get('domain_vars', None))


def read_state(params: dict, domain: xr.Dataset) -> xr.Dataset:
    """Load in a state file"""
    read_funcs = {
        "netcdf": read_netcdf,
        "data": read_data
    }

    return read_funcs[params['state_fmt']](
        params['state'], domain=domain, iter_dims=params['iter_dims'],
        start=params['state_start'], stop=params['state_stop'],
        calendar=params['calendar'],
        var_dict=params.get('state_vars', None))


def process_nc(params: dict, domain: xr.Dataset) -> xr.Dataset:
    """Process NetCDF-like Data"""
    read_funcs = {
        "netcdf": read_netcdf,
        "data": read_data
    }
    return read_funcs[params['forcing_fmt']](
        params['forcing'], domain=domain, iter_dims=params['iter_dims'],
        start=params['start'], stop=params['stop'],
        calendar=params['calendar'], var_dict=params.get('forcing_vars', None))


def process_vic(params: dict, domain: xr.Dataset) -> xr.Dataset:
    """Process VIC-like data"""
    read_funcs = {
        "binary": read_binary,
        "ascii": read_ascii,
    }

    if 'lon' not in params['iter_dims'] or 'lat' not in params['iter_dims']:
        raise ValueError(
            'Using VIC type input requires lat and lon to be'
            ' specified via `iter_dims` in configuration.')

    # Creates the master dataset which will be used to parallelize
    dates = date_range(params['start'], params['stop'],
                       calendar=params['calendar'])
    coords = {'time': dates,
              'lon': domain['lon'],
              'lat': domain['lat']}
    shape = (len(dates), len(domain['lat']), len(domain['lon']))
    dims = ('time', 'lat', 'lon', )

    met_data = xr.Dataset(
        coords=coords, attrs={'n_days': len(dates)})
    for var in params['forcing_vars']:
        met_data[var] = xr.DataArray(data=np.full(shape, np.nan),
                                     coords=coords, dims=dims, name=var)

    # Fill in the data
    for job in params['forcing']:
        try:
            _, lat, lon = os.path.basename(job).split("_")[-3:]
            lat, lon = float(lat), float(lon)
            if not domain['mask'].sel(lat=lat, lon=lon).values > 0:
                continue
            ds = read_funcs[params['forcing_fmt']](
                job, start=params['start'], stop=params['stop'],
                calendar=params['calendar'], var_dict=params['forcing_vars'])
            for var in params['forcing_vars'].keys():
                met_data[var].loc[{'lat': lat, 'lon': lon}] = ds[var]
        except (ValueError, KeyError):
            continue
    return met_data


def read_ascii(data_handle, domain=None, iter_dims=['lat', 'lon'],
               start=None, stop=None, calendar='standard',
               var_dict=None) -> xr.Dataset:
    """Read in an ascii forcing file"""
    dates = date_range(start, stop, calendar=calendar)
    names = var_dict.keys()
    ds = pd.read_table(data_handle, header=None, delim_whitespace=True,
                       names=names).head(len(dates))
    ds.index = dates
    return ds


def read_netcdf(data_handle, domain=None, iter_dims=['lat', 'lon'],
                start=None, stop=None, calendar='standard',
                var_dict=None) -> xr.Dataset:
    """Read in a NetCDF file"""
    ds = xr.open_mfdataset(data_handle)

    if 'time' in ds.coords:
        ds['time'] = (ds.indexes['time'] -
                      pd.Timedelta('11H59M59S')).round('D')

    if var_dict is not None:
        var_list = list(var_dict.keys())
        ds = ds[var_list]
        ds.rename(var_dict, inplace=True)

    if start is not None or stop is not None:
        ds = ds.sel(time=slice(start, stop))
        dates = ds.indexes['time']
        ds['day_of_year'] = xr.Variable(('time', ), dates.dayofyear)

    if domain is not None:
        ds = ds.sel(**{d: domain[d] for d in iter_dims})
    return ds


def read_data(data_handle, domain=None, iter_dims=['lat', 'lon'],
              start=None, stop=None, calendar='standard',
              var_dict=None) -> xr.Dataset:
    """Read data directly from an xarray dataset"""
    varlist = list(data_handle.keys())
    if var_dict is not None:
        data_handle.rename(var_dict, inplace=True)
        varlist = list(var_dict.values())
    data_handle = data_handle[varlist]

    if start is not None and stop is not None:
        data_handle = data_handle.sel(time=slice(start, stop))
        dates = data_handle.indexes['time']
        data_handle['day_of_year'] = xr.Variable(('time', ), dates.dayofyear)

    if domain is not None:
        data_handle = data_handle.sel(**{d: domain[d] for d in iter_dims})
    return data_handle


def read_binary(data_handle, domain=None, iter_dims=['lat', 'lon'],
                start=None, stop=None, calendar='standard',
                var_dict=None) -> xr.Dataset:
    """Reads a binary forcing file (VIC 4 format)"""
    dates = date_range(start, stop, calendar=calendar)
    n_days = len(dates)
    type_lookup = {'signed': 'h', 'unsigned': 'H'}
    # Pack these for nicer syntax in the loop
    var_names = var_dict.keys()
    data_list = [[] for var in var_dict.keys()]
    n_vars = len(var_names)
    params = [p for p in var_dict.values()]
    scales = [float(s.split()[0]) for s in params]
    datatypes = [type_lookup[s.split()[-1]] for s in params]
    with open(data_handle, 'rb') as f:
        i = 0
        points_read = 0
        points_needed = 4 * n_days
        while points_read != points_needed:
            bytes = f.read(2)
            if bytes:
                # Get correct variable and data type with i,
                # then unpack & scale
                data_list[i].append(
                    struct.unpack(datatypes[i], bytes)[0] / scales[i])
                i = (i + 1) % n_vars
                points_read += 1
            else:
                break

    # Binary forcing files have naming format $NAME_$LAT_$LON
    param_list = os.path.basename(data_handle).split("_")[-3:]
    params = {"name": param_list[0],
              "lat": float(param_list[1]),
              "lon": float(param_list[2]),
              "n_days": int(n_days)}

    # Assemble the dataset
    data_dict = {c[0]: (['time'], c[1]) for c in zip(var_names, data_list)}
    data_dict['day_of_year'] = (['time'], dates.dayofyear)
    df = xr.Dataset(data_dict,
                    coords={'lon': [params['lon']],
                            'lat': [params['lat']],
                            'time': dates},
                    attrs={'n_days': params['n_days']})
    return df



"""
MTCLIM
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import metsim.constants as cnst
from metsim.physics import atm_pres, calc_pet, svp


def run(df, params):
    df['t_day'] = t_day(df['t_min'].values, df['t_max'].values, params)
    df['tfmax'] = tfmax(df['dtr'].values, df['smoothed_dtr'].values,
                        df['prec'].values, params)
    df['tskc'] = tskc(df['tfmax'].values, params)

    tdew_old = df['t_min'].values
    vp_temp = vapor_pressure(df['t_min'].values)
    sw_temp = shortwave(df['tfmax'].values, vp_temp,
                        df['tt_max'], df['potrad'].values)
    pet_temp = pet(sw_temp, df['t_day'].values, df['daylength'].values, params)
    tdew_temp = tdew(pet_temp, df['t_min'].values, df['seasonal_prec'].values,
                     df['dtr'].values)

    while(np.sqrt(np.mean((tdew_temp - tdew_old)**2)) > params['tdew_tol']):
        tdew_old = tdew_temp.copy()
        vp_temp = vapor_pressure(tdew_temp)
        sw_temp = shortwave(df['tfmax'].values, vp_temp,
                            df['tt_max'].values, df['potrad'].values)
        pet_temp = pet(sw_temp, df['t_day'].values, df['daylength'].values,
                       params)
        tdew_temp = tdew(pet_temp, df['t_min'].values,
                         df['seasonal_prec'].values, df['dtr'].values)

    df['tdew'] = tdew_temp
    df['vapor_pressure'] = vapor_pressure(df['tdew'].values)
    df['shortwave'] = shortwave(df['tfmax'].values,
                                df['vapor_pressure'].values,
                                df['tt_max'].values, df['potrad'].values)
    df['pet'] = pet(df['shortwave'].values, df['t_day'].values,
                    df['daylength'].values, params)
    return df


def t_day(t_min, t_max, params):
    t_mean = (t_min + t_max) / 2
    return ((t_max - t_mean) * params['tday_coef']) + t_mean


def tfmax(dtr, sm_dtr, prec, params):
    b = cnst.B0 + cnst.B1 * np.exp(-cnst.B2 * sm_dtr)
    tfmax = 1.0 - 0.9 * np.exp(-b * np.power(dtr, cnst.C))
    inds = np.array(prec > params['sw_prec_thresh'])
    tfmax[inds] *= params['rain_scalar']
    return tfmax


def pet(shortwave, t_day, daylength, params):
    pa = atm_pres(params['elev'], params['lapse_rate'])
    return calc_pet(shortwave, t_day, daylength, pa) * cnst.MM_PER_CM


def tdew(pet, t_min, seasonal_prec, dtr):
    parray = seasonal_prec < 80.0
    seasonal_prec[parray] = 80.0
    ratio = pet / seasonal_prec
    return ((t_min + cnst.KELVIN) * (-0.127 + 1.121 * (1.003 - 1.444 * ratio + 12.312 * \
            np.power(ratio, 2) - 32.766 * np.power(ratio, 3)) + 0.0006 * dtr) - cnst.KELVIN)


def vapor_pressure(tdew):
    return svp(tdew)


def shortwave(tfmax, vapor_pressure, tt_max, potrad):
    t_tmax = np.maximum(tt_max + (cnst.ABASE * vapor_pressure), 0.0001)
    return potrad * t_tmax * tfmax


def tskc(tfmax, params):
    if (params['lw_cloud'].upper() == 'CLOUD_DEARDORFF'):
        return 1. - tfmax
    return np.sqrt((1. - tfmax) / 0.65)



""" Disaggregates daily data down to finer grained data using some heuristics
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
from typing import Tuple

import numpy as np
import pandas as pd
import scipy.interpolate

import metsim.constants as cnst
from metsim.datetime import date_range
from metsim.physics import svp


def disaggregate(df_daily: pd.DataFrame, params: dict,
                 solar_geom: dict, t_begin: list=None,
                 t_end: list=None) -> pd.DataFrame:
    """
    Take a daily timeseries and scale it down to a finer
    time scale.

    Parameters
    ----------
    df_daily:
        Dataframe containing daily timeseries.
        Should be the result of one of the methods
        provided in the `methods` directory.
    params:
        A dictionary containing the class parameters
        of the MetSim object.
    solar_geom:
        A dictionary of solar geometry variables
    t_begin:
        List of t_min and t_max for day previous to the
        start of `df_daily`. None indicates no extension
        of the record.
    t_end:
        List of t_min and t_max for day after the end
        of `df_daily`. None indicates no extension of
        the record.

    Returns
    -------
    df_disagg:
        A dataframe with sub-daily timeseries.
    """
    stop = (df_daily.index[-1] + pd.Timedelta('1 days') -
            pd.Timedelta("{} minutes".format(params['time_step'])))
    dates_disagg = date_range(df_daily.index[0], stop,
                              freq='{}T'.format(params['time_step']))
    df_disagg = pd.DataFrame(index=dates_disagg)
    n_days = len(df_daily)
    n_disagg = len(df_disagg)
    ts = int(params['time_step'])
    df_disagg['shortwave'] = shortwave(df_daily['shortwave'].values,
                                       df_daily['daylength'].values,
                                       df_daily.index.dayofyear.values,
                                       solar_geom['tiny_rad_fract'],
                                       params)

    t_Tmin, t_Tmax = set_min_max_hour(solar_geom['tiny_rad_fract'],
                                      df_daily.index.dayofyear.values - 1,
                                      n_days, ts, params)

    df_disagg['temp'] = temp(
        df_daily['t_min'].values, df_daily['t_max'].values,
        n_disagg, t_Tmin, t_Tmax, ts, t_begin, t_end)

    df_disagg['vapor_pressure'] = vapor_pressure(
        df_daily['vapor_pressure'].values, df_disagg['temp'].values,
        t_Tmin, n_disagg, ts)

    df_disagg['rel_humid'] = relative_humidity(
        df_disagg['vapor_pressure'].values, df_disagg['temp'].values)

    df_disagg['air_pressure'] = pressure(df_disagg['temp'].values,
                                         params['elev'], params['lapse_rate'])

    df_disagg['spec_humid'] = specific_humidity(
        df_disagg['vapor_pressure'].values,
        df_disagg['air_pressure'].values)

    df_disagg['tskc'] = tskc(df_daily['tskc'].values, ts)

    df_disagg['longwave'] = longwave(
        df_disagg['temp'].values, df_disagg['vapor_pressure'].values,
        df_disagg['tskc'].values, params)

    df_disagg['prec'] = prec(df_daily['prec'].values, ts)

    if 'wind' in df_daily:
        df_disagg['wind'] = wind(df_daily['wind'].values, ts)

    return df_disagg.fillna(method='ffill')


def set_min_max_hour(tiny_rad_fract: np.array, yday: np.array, n_days: int,
                     ts: int, params: dict) -> Tuple[np.array]:
    """
    Determine the time at which min and max temp
    is reached for each day.

    Parameters
    ----------
    tiny_rad_fract:
        Array of fraction of shortwave radiation received
        at a shortened timestep. This should be calculated
        by `metsim.physics.solar_geom`.
    yday:
        Array of day of year for each simulated day.
    n_days:
        Number of days in run.
    ts:
        Timestep of run.
    params:
        Dictionary of parameters to use.  Must contain
        `utc_offset` and `tmax_daylength_fraction`.

    Returns
    -------
    (t_t_min, t_t_max):
        A tuple containing 2 timeseries, corresponding
        to time of min and max temp, respectively
    """

    # calculate minute of sunrise and sunset for each day of the year
    rad_mask = 1 * (tiny_rad_fract > 0)
    mask = np.diff(rad_mask)
    rise_times = np.where(mask > 0)[1] * (cnst.SW_RAD_DT / cnst.SEC_PER_MIN)
    set_times = np.where(mask < 0)[1] * (cnst.SW_RAD_DT / cnst.SEC_PER_MIN)

    if params['utc_offset']:
        # not used elsewhere:
        # rad_fract_per_day = int(cnst.SEC_PER_DAY/cnst.SW_RAD_DT)
        utc_offset = int(((params.get("lon", 0) - params.get("theta_s", 0)) /
                          cnst.DEG_PER_REV) * cnst.MIN_PER_DAY)
        rise_times -= utc_offset
        set_times -= utc_offset

    # map the daily sunrise and sunset to a monotonic timseries (in minutes)
    offset = np.arange(0, n_days * cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY,
                       cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY)
    rise_times = rise_times[yday] + offset
    set_times = set_times[yday] + offset

    # time of maximum and minimum temperature calculated thusly
    t_t_max = (params['tmax_daylength_fraction'] * (set_times - rise_times) +
               rise_times) + ts
    t_t_min = rise_times
    return t_t_min, t_t_max


def temp(t_min: np.array, t_max: np.array, out_len: int,
         t_t_min: np.array, t_t_max: np.array, ts: int,
         t_begin: list=None, t_end: list=None) -> np.array:
    """
    Disaggregate temperature using a Hermite polynomial
    interpolation scheme.

    Parameters
    ----------
    t_min:
        Timeseries of daily minimum temperatures.
    t_max:
        Timeseries of daily maximum temperatures.
    out_len:
        Length of the required output vector.
    t_t_min:
        Times at which minimum daily
        temperatures are reached.
    t_t_max:
        Times at which maximum daily
        temperatures are reached.
    ts:
        Timestep for disaggregation
    t_begin:
        List of t_min and t_max for day previous to the
        start of `df_daily`. None indicates no extension
        of the record.
    t_end:
        List of t_min and t_max for day after the end
        of `df_daily`. None indicates no extension of
        the record.

    Returns
    -------
    temps:
        A sub-daily timeseries of temperature.
    """
    # Calculate times of min/max temps
    time = np.array(list(next(it) for it in itertools.cycle(
        [iter(t_t_min), iter(t_t_max)])))
    temp = np.array(list(next(it) for it in itertools.cycle(
        [iter(t_min), iter(t_max)])))
    # Account for end points
    ts_ends = cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY
    time = np.append(np.insert(time, 0, time[0:2] - ts_ends),
                     time[-2:] + ts_ends)

    # If no start or end data is provided to extend the record repeat values
    # This provides space at the ends so that extrapolation doesn't continue
    # in strange ways at the end points
    if t_begin is None:
        t_begin = temp[0:2]
    if t_end is None:
        t_end = temp[-2:]
    temp = np.append(np.insert(temp, 0, t_begin), t_end)

    # Interpolate the values
    interp = scipy.interpolate.PchipInterpolator(time, temp, extrapolate=True)
    temps = interp(ts * np.arange(0, out_len))
    return temps


def prec(prec: np.array, ts: int) -> np.array:
    """
    Splits the daily precipitation evenly throughout the day
    Note: this returns only through to the beginning of the
          last day.  Final values are filled in using a
          forward fill in the top level disaggregate function

    Parameters
    ----------
    prec:
        Daily timeseries of precipitation
    ts:
        Timestep to disaggregate down to

    Returns
    -------
    prec:
        A sub-daily timeseries of precipitation
    """
    scale = ts / (cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY)
    n_repeats = 1 / scale
    return np.repeat(prec * scale, n_repeats)


def wind(wind: np.array, ts: int) -> np.array:
    """
    Wind is assumed constant throughout the day
    Note: this returns only through to the beginning of the
          last day.  Final values are filled in using a
          forward fill in the top level disaggregate function

    Parameters
    ----------
    wind:
        Daily timeseries of wind
    ts:
        Timestep to disaggregate down to

    Returns
    -------
    wind:
        A sub-daily timeseries of wind
    """
    n_repeats = (cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY) / ts
    return np.repeat(wind, n_repeats)


def pressure(temp: np.array, elev: float, lr: float) -> np.array:
    """
    Calculates air pressure.

    Parameters
    ----------
    temp:
        A sub-daily timeseries of temperature
    elev:
        Elevation
    lr:
        Lapse rate

    Returns
    -------
    pressure:
        A sub-daily timeseries of air pressure (kPa)
    """
    temp_corr = cnst.KELVIN + temp + 0.5 * elev * -lr
    ratio = -(elev * cnst.G_STD) / (cnst.R_DRY * temp_corr)
    return cnst.P_STD * np.exp(ratio) / cnst.MBAR_PER_BAR


def specific_humidity(vapor_pressure: np.array,
                      air_pressure: np.array) -> np.array:
    """
    Calculates specific humidity

    Parameters
    ----------
    vapor_pressure:
        A sub-daily timeseries of vapor pressure (kPa)
    air_pressure:
        A sub-daily timeseries of air pressure (kPa)

    Returns
    -------
    spec_humid:
        A sub-daily timeseries of specific humidity
    """
    mix_rat = (cnst.EPS * vapor_pressure) / (air_pressure - vapor_pressure)
    return mix_rat / (1 + mix_rat)


def relative_humidity(vapor_pressure: np.array,
                      temp: np.array) -> np.array:
    """
    Calculate relative humidity from vapor pressure
    and temperature.

    Parameters
    ----------
    vapor_pressure:
        A sub-daily timeseries of vapor pressure
    temp:
        A sub-daily timeseries of temperature

    Returns
    -------
    rh:
        A sub-daily timeseries of relative humidity
    """
    rh = (cnst.MAX_PERCENT * cnst.MBAR_PER_BAR * (vapor_pressure / svp(temp)))
    rh[rh > cnst.MAX_PERCENT] = cnst.MAX_PERCENT
    return rh


def vapor_pressure(vp_daily: np.array, temp: np.array, t_t_min: np.array,
                   n_out: int, ts: int) -> np.array:
    """
    Calculate vapor pressure.  First a linear interpolation
    of the daily values is calculated.  Then this is compared
    to the saturated vapor pressure calculated using the
    disaggregated temperature. When the interpolated vapor
    pressure is greater than the calculated saturated
    vapor pressure, the interpolation is replaced with the
    saturation value.

    Parameters
    ----------
    vp_daily:
        Daily vapor pressure
    temp:
        Sub-daily temperature
    t_t_min:
        Timeseries of minimum daily temperature
    n_out:
        Number of output observations
    ts:
        Timestep to disaggregate down to

    Returns
    -------
    vp:
        A sub-daily timeseries of the vapor pressure
    """
    # Linearly interpolate the values
    interp = scipy.interpolate.interp1d(t_t_min, vp_daily / cnst.MBAR_PER_BAR,
                                        fill_value='extrapolate')
    vp_disagg = interp(ts * np.arange(0, n_out))

    # Account for situations where vapor pressure is higher than
    # saturation point
    vp_sat = svp(temp) / cnst.MBAR_PER_BAR
    vp_disagg = np.where(vp_sat < vp_disagg, vp_sat, vp_disagg)
    return vp_disagg


def longwave(air_temp: np.array, vapor_pressure: np.array,
             tskc: np.array, params: dict) -> np.array:
    """
    Calculate longwave. This calculation can be performed
    using a variety of parameterizations for both the
    clear sky and cloud covered emissivity. Options for
    choosing these parameterizations should be passed in
    via the `params` argument.

    For more information about the options provided in this
    function see:

    .. [1] Bohn, T.J., Livneh, B., Oyler, J.W., Running, S.W., Nijssen, B.
      and Lettenmaier, D.P., 2013. Global evaluation of MTCLIM and
      related algorithms for forcing of ecological and hydrological
      models.  Agricultural and forest meteorology, 176, pp.38-49,
      doi:10.1016/j.agrformet.2013.03.003.

    Parameters
    ----------
    air_temp:
        Sub-daily temperature
    vapor_pressure:
        Sub-daily vapor pressure
    tskc:
        Daily cloud fraction
    params:
        A dictionary of parameters, which contains
        information about which emissivity and cloud
        fraction methods to use.

    Returns
    -------
    lwrad:
        A sub-daily timeseries of the longwave radiation
    """
    emissivity_calc = {
        # TVA 1972
        # Tennessee Valley Authority, 1972. Heat and mass transfer between a
        # water surface and the atmosphere. Tennessee Valley Authority, Norris,
        # TN. Laboratory report no. 14. Water resources research report
        # no. 0-6803.
        'TVA': lambda vp: 0.74 + 0.0049 * vp,
        # Anderson 1954
        # Anderson, E.R., 1954. Energy budget studies, water loss
        # investigations: lake Hefner studies. U.S. Geol. Surv. Prof. Pap. 269,
        # 71–119 [Available from U.S. Geological Survey, 807 National Center,
        # Reston, VA 20192.].
        'ANDERSON': lambda vp: 0.68 + 0.036 * np.power(vp, 0.5),
        # Brutsaert 1975
        # Brutsaert, W., 1975. On a derivable formula for long-wave radiation
        # from clear skies. Water Resour. Res. 11 (5), 742–744,
        # doi:10.1029/WR011i005p00742.
        'BRUTSAERT': lambda vp: 1.24 * np.power(vp / air_temp, 0.14285714),
        # Satterlund 1979
        # Satterlund, D.R., 1979. An improved equation for estimating long-wave
        # radiation from the atmosphere. Water Resour. Res. 15 (6), 1649–1650,
        # doi:10.1029/WR015i006p01649.
        'SATTERLUND': lambda vp: 1.08 * (
            1 - np.exp(-1 * np.power(vp, (air_temp / 2016)))),
        # Idso 1981
        # Idso, S.B., 1981. A set of equations for full spectrum and 8- to
        # 14-µm and 10.5- to 12.5-µm, thermal radiation from cloudless skies.
        # Water Resour. Res. 17 (2), 295–304, doi:10.1029/WR017i002p00295.
        'IDSO': lambda vp: 0.7 + 5.95e-5 * vp * np.exp(1500 / air_temp),
        # Prata 1996
        # Prata, A.J., 1996. A new long-wave formula for estimating downward
        # clear-sky radiation at the surface. Q. J. R. Meteor. Soc. 122 (533),
        # 1127–1151, doi:10.1002/qj.49712253306.
        'PRATA': lambda vp: (1 - (1 + (46.5 * vp / air_temp)) * np.exp(
            -np.sqrt((1.2 + 3. * (46.5 * vp / air_temp)))))}
    cloud_calc = {
        # TVA 1972 (see above)
        'TVA': lambda emis, tskc: (1.0 + (0.17 * tskc ** 2)) * emis,
        # Deardorff 1978
        # Deardorff, J.W., 1978. Efficient prediction of ground surface
        # temperature and moisture, with an inclusion of a layer of vegetation.
        # J. Geophys. Res. 83 (N64), 1889–1903, doi:10.1029/JC083iC04p01889.
        'CLOUD_DEARDORFF': lambda emis, tskc: tskc + (1 - tskc) * emis
    }
    # Re-index and fill cloud cover, then convert temps to K
    air_temp = air_temp + cnst.KELVIN
    vapor_pressure = vapor_pressure * 10

    # Calculate longwave radiation based on the options
    emiss_func = emissivity_calc[params['lw_type'].upper()]
    emissivity_clear = emiss_func(vapor_pressure)
    emiss_func = cloud_calc[params['lw_cloud'].upper()]
    emissivity = emiss_func(emissivity_clear, tskc)
    lwrad = emissivity * cnst.STEFAN_B * np.power(air_temp, 4)
    return lwrad


def tskc(tskc: np.array, ts: int) -> np.array:
    """
    Disaggregate cloud fraction with uniform interpolation

    Parameters
    ----------
    tskc:
        Daily cloud fraction
    ts:
        Time step to disaggregate to (in minutes)

    Returns
    -------
    tskc:
        Sub-daily timeseries of cloud fraction
    """
    n_repeats = (cnst.MIN_PER_HOUR * cnst.HOURS_PER_DAY) / ts
    return np.repeat(tskc, n_repeats)


def shortwave(sw_rad: np.array, daylength: np.array, day_of_year: np.array,
              tiny_rad_fract: np.array, params: dict) -> np.array:
    """
    Disaggregate shortwave radiation down to a subdaily timeseries.

    Parameters
    ----------
    sw_rad:
        Daily incoming shortwave radiation
    daylength:
        List of daylength time for each day of year
    day_of_year:
        Timeseries of index of days since Jan-1
    tiny_rad_fract:
        Fraction of the daily potential radiation
        during a radiation time step defined by SW_RAD_DT
    params:
        Dictionary of parameters from the MetSim object

    Returns
    -------
    disaggrad:
        A sub-daily timeseries of shortwave radiation.
    """
    ts = int(params['time_step'])
    ts_hourly = float(ts) / cnst.MIN_PER_HOUR
    tmp_rad = (sw_rad * daylength) / (cnst.SEC_PER_HOUR * ts_hourly)
    n_days = len(tmp_rad)
    ts_per_day = int(cnst.HOURS_PER_DAY * cnst.MIN_PER_HOUR / ts)
    disaggrad = np.zeros(int(n_days * ts_per_day))
    rad_fract_per_day = int(cnst.SEC_PER_DAY / cnst.SW_RAD_DT)
    if params['utc_offset']:
        utc_offset = int(((params.get("lon", 0) - params.get("theta_s", 0)) /
                          cnst.DEG_PER_REV) * rad_fract_per_day)
        tiny_rad_fract = np.roll(tiny_rad_fract.flatten(), -utc_offset)
    else:
        tiny_rad_fract = tiny_rad_fract.flatten()
    chunk_size = int(ts * (cnst.SEC_PER_MIN / cnst.SW_RAD_DT))
    ts_id = np.repeat(np.arange(ts_per_day), chunk_size)
    for day in range(n_days):
        radslice = slice((day_of_year[day] - 1) * rad_fract_per_day,
                         (day_of_year[day]) * rad_fract_per_day)
        rad = tiny_rad_fract[radslice]
        dslice = slice(int(day * ts_per_day), int((day + 1) * ts_per_day))
        rad_chunk = np.bincount(ts_id, weights=rad)
        disaggrad[dslice] = rad_chunk * tmp_rad[day]
    return disaggrad



'''
physics
'''
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas as pd
from numba import jit

import metsim.constants as cnst


def calc_pet(rad: np.array, ta: np.array, dayl: np.array,
             pa: float, dt: float=0.2) -> np.array:
    '''
    Calculates the potential evapotranspiration for aridity corrections in
    `calc_vpd()`, according to Kimball et al., 1997

    Parameters
    ----------
    rad:
        daylight average incident shortwave radiation (W/m2)
    ta:
        daylight average air temperature (deg C)
    dayl:
        daylength (s)
    pa:
        air pressure (Pa)
    dt:
        offset for saturation vapor pressure calculation

    Returns
    -------
    pet
        Potential evapotranspiration (cm/day)
    '''
    # Definition of parameters:
    # rnet       # (W m-2) absorbed shortwave radiation avail. for ET
    # lhvap      # (J kg-1) latent heat of vaporization of water
    # gamma      # (Pa K-1) psychrometer parameter
    # dt = 0.2   # offset for saturation vapor pressure calculation
    # t1, t2     # (deg C) air temperatures
    # pvs1, pvs2 # (Pa)   saturated vapor pressures
    # pet        # (kg m-2 day-1) potential evapotranspiration
    # s          # (Pa K-1) slope of saturated vapor pressure curve

    # calculate absorbed radiation, assuming albedo = 0.2  and ground
    # heat flux = 10% of absorbed radiation during daylight
    rnet = rad * 0.72

    # calculate latent heat of vaporization as a function of ta
    lhvap = 2.5023e6 - 2430.54 * ta

    # calculate the psychrometer parameter: gamma = (cp pa)/(lhvap epsilon)
    # where:
    # cp       (J/kg K)   specific heat of air
    # epsilon  (unitless) ratio of molecular weights of water and air
    gamma = cnst.CP * pa / (lhvap * cnst.EPS)

    # estimate the slope of the saturation vapor pressure curve at ta
    # temperature offsets for slope estimate
    t1 = ta + dt
    t2 = ta - dt

    # calculate saturation vapor pressures at t1 and t2, using formula from
    # Abbott, P.F., and R.C. Tabony, 1985. The estimation of humidity
    # parameters. Meteorol. Mag., 114:49-56.
    pvs1 = svp(t1)
    pvs2 = svp(t2)

    # calculate slope of pvs vs. T curve near ta
    s = (pvs1 - pvs2) / (t1 - t2)
    # can this be s = svp_slope(ta)? JJH

    # calculate PET using Priestly-Taylor approximation, with coefficient
    # set at 1.26. Units of result are kg/m^2/day, equivalent to mm water/day
    pet = (1.26 * (s / (s + gamma)) * rnet * dayl) / lhvap

    # return a value in centimeters/day, because this value is used in a ratio
    # to annual total precip, and precip units are centimeters
    return (pet / 10.)


def atm_pres(elev: float, lr: float) -> float:
    '''
    Atmospheric pressure (Pa) as a function of elevation (m)

    .. [1] Iribane, J.V., and W.L. Godson, 1981. Atmospheric
      Thermodynamics, 2nd Edition. D. Reidel Publishing
      Company, Dordrecht, The Netherlands (p. 168)

    Parameters
    ----------
    elev:
        Elevation in meters
    lr:
        Lapse rate (K/m)

    Returns
    -------
    pressure:
        Atmospheric pressure (Pa)
    '''
    t1 = 1.0 - (lr * elev) / cnst.T_STD
    t2 = cnst.G_STD / (lr * (cnst.R / cnst.MA))
    return cnst.P_STD * np.power(t1, t2)


def svp(temp: np.array, a: float=0.61078, b: float=17.269, c: float=237.3):
    '''
    Compute the saturated vapor pressure.

    .. [2] Maidment, David R. Handbook of hydrology. McGraw-Hill Inc.,
      1992 Equation 4.2.2.

    Parameters
    ----------
    temp:
        Temperature (degrees Celsius)
    a:
        (optional) parameter
    b:
        (optional) parameter
    c:
        (optional) parameter

    Returns
    -------
    svp:
        Saturated vapor pressure (Pa)
    '''
    svp = a * np.exp((b * temp) / (c + temp))
    inds = np.nonzero(temp < 0.)[0]
    svp[inds] *= 1.0 + .00972 * temp[inds] + .000042 * np.power(temp[inds], 2)
    return svp * 1000.


def svp_slope(temp: pd.Series, a: float=0.61078,
              b: float=17.269, c: float=237.3):
    '''
    Compute the gradient of the saturated vapor pressure as a function of
    temperature.

    .. [3] Maidment, David R. Handbook of hydrology. McGraw-Hill Inc.,
      1992. Equation 4.2.3.

    Parameters
    ----------
    temp:
        Temperature (degrees Celsius)

    Returns
    -------
    dsvp_dT:
        Gradient of d(svp)/dT.
    '''
    return (b * c) / ((c + temp) * (c + temp)) * svp(temp, a=a, b=b, c=c)


@jit(nopython=True)
def solar_geom(elev: float, lat: float, lr: float) -> tuple:
    """
    Flat earth assumption

    Parameters
    ----------
    elev:
        Elevation in meters
    lat:
        Latitude in decimal format
    lr:
        Lapse rate in K/m

    Returns
    -------
    sg:
        (tiny_rad_fract, daylength, flat_potrad, tt_max0)
    """
    # optical airmass by degrees
    OPTAM = [2.90, 3.05, 3.21, 3.39, 3.69, 3.82, 4.07,
             4.37, 4.72, 5.12, 5.60, 6.18, 6.88, 7.77,
             8.90, 10.39, 12.44, 15.36, 19.79, 26.96, 30.00]
    dayperyear = int(np.ceil(cnst.DAYS_PER_YEAR))
    tt_max0 = np.zeros(dayperyear)
    daylength = np.zeros(dayperyear)
    flat_potrad = np.zeros(dayperyear)

    # Calculate pressure ratio as a function of elevation
    t1 = 1.0 - (lr * elev) / cnst.T_STD
    t2 = cnst.G_STD / (lr * (cnst.R / cnst.MA))
    trans = np.power(cnst.TBASE, np.power(t1, t2))

    # Translate lat to rad
    lat = np.minimum(
        np.maximum(
            lat * cnst.RAD_PER_DEG, -np.pi / 2.), np.pi / 2.0)
    coslat = np.cos(lat)
    sinlat = np.sin(lat)

    # Sub-daily time step and angular step
    dt = cnst.SW_RAD_DT
    dh = dt / cnst.SEC_PER_RAD

    # Allocate the radiation arrays
    tiny_step_per_day = int(cnst.SEC_PER_DAY / cnst.SW_RAD_DT)
    tiny_rad_fract = np.zeros((dayperyear, tiny_step_per_day))
    for i in range(dayperyear - 1):
        # Declination and quantities of interest
        decl = cnst.MIN_DECL * np.cos((i + cnst.DAYS_OFF) * cnst.RAD_PER_DAY)
        cosdecl = np.cos(decl)
        sindecl = np.sin(decl)

        # calculate daylength as a function of lat and decl
        cosegeom = coslat * cosdecl
        sinegeom = sinlat * sindecl
        coshss = min(max(-sinegeom / cosegeom, -1), 1)
        hss = np.arccos(coshss)
        daylength[i] = min(2.0 * hss * cnst.SEC_PER_RAD, cnst.SEC_PER_DAY)
        # Extraterrestrial radiation perpendicular to beam,
        # total over the timestep (J)
        dir_beam_topa = (1368.0 + 45.5 * np.sin(
            (2.0 * np.pi * i / cnst.DAYS_PER_YEAR) + 1.7)) * dt
        sum_trans = 0
        sum_flat_potrad = 0
        # Set up angular calculations
        for h in np.arange(-hss, hss, dh):
            # Cosine of the hour angle and solar zenith angle
            cosh = np.cos(h)
            cza = cosegeom * cosh + sinegeom
            if (cza > 0):
                # When sun is above flat horizon do flat-surface
                # calculations to determine daily total transmittance
                # and save potential radiation for calculation of
                # diffuse portion
                dir_flat_topa = dir_beam_topa * cza
                am = 1.0 / (cza + 0.0000001)
                if (am > 2.9):
                    ami = min(max(int(
                        np.arccos(cza) / cnst.RAD_PER_DEG) - 69, 0), 20)
                    am = OPTAM[ami]
                sum_trans += (np.power(trans, am) * dir_flat_topa)
                sum_flat_potrad += dir_flat_topa
            else:
                # Sun not above horizon
                dir_flat_topa = 0

            tinystep = int(min(max(
                (12 * cnst.SEC_PER_HOUR + h * cnst.SEC_PER_RAD) / dt, 0),
                tiny_step_per_day - 1))
            tiny_rad_fract[i][tinystep] = dir_flat_topa

        if daylength[i] and sum_flat_potrad > 0:
            tiny_rad_fract[i] /= sum_flat_potrad

        if daylength[i]:
            # Transmittance and potential radiation
            # averaged over daylength
            tt_max0[i] = sum_trans / sum_flat_potrad
            flat_potrad[i] = sum_flat_potrad / daylength[i]
        else:
            # No daytime - no radiation
            tt_max0[i] = 0.
            flat_potrad[i] = 0.
    tt_max0[dayperyear - 1] = tt_max0[dayperyear - 2]
    flat_potrad[dayperyear - 1] = flat_potrad[dayperyear - 2]
    daylength[dayperyear - 1] = daylength[dayperyear - 2]
    tiny_rad_fract[dayperyear - 1] = tiny_rad_fract[dayperyear - 2]
    return tiny_rad_fract, daylength, flat_potrad, tt_max0



#!/usr/bin/env python

from setuptools import setup

import versioneer

setup(name='metsim',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Meteorology Simulator',
      url='https://github.com/UW-Hydro/MetSim',
      author='Andrew Bennett',
      author_email='bennett.andr@gmail.com',
      packages=['metsim', 'metsim.methods', 'metsim.cli'],
      entry_points={
          'console_scripts': ['ms = metsim.cli.ms:main']},
      install_requires=['xarray', 'numba', 'numpy', 'pandas',
                        'dask', 'toolz', 'netCDF4', 'scipy'],
      keywords=['meteorology', 'disaggregation', 'hydrology',
                'climate', 'mtclim'],
      tests_require=['pytest'],)



"""
The main object of the MetSim package. The MetSim object
is used to set up and launch forcing generation and/or
disaggregation routines.

The MetSim object uses a class dictionary to refer to
the model setup, which can be modified after instantiation
if necessary.  Before calling `run` or `launch` on the
instance it is required to call the `load` function to
ensure that all of the required parameters have been
set and that the input data is sufficient to provide the
output specified.
"""
# Meteorology Simulator
# Copyright (C) 2017  The Computational Hydrology Group, Department of Civil
# and Environmental Engineering, University of Washington.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import json
import logging
import os
import sys
import time as tm
from collections import Iterable, OrderedDict
from getpass import getuser
from multiprocessing import Pool

import numpy as np
import pandas as pd
import xarray as xr

import metsim.constants as cnst
from metsim import io
from metsim.datetime import date_range
from metsim.disaggregate import disaggregate
from metsim.methods import mtclim
from metsim.physics import solar_geom

references = '''Thornton, P.E., and S.W. Running, 1999. An improved algorithm for estimating incident daily solar radiation from measurements of temperature, humidity, and precipitation. Agricultural and Forest Meteorology, 93:211-228.
Kimball, J.S., S.W. Running, and R. Nemani, 1997. An improved method for estimating surface humidity from daily minimum temperature. Agricultural and Forest Meteorology, 85:87-98.
Bohn, T. J., B. Livneh, J. W. Oyler, S. W. Running, B. Nijssen, and D. P. Lettenmaier, 2013a: Global evaluation of MT-CLIM and related algorithms for forcing of ecological and hydrological models, Agr. Forest. Meteorol., 176, 38-49, doi:10.1016/j.agrformet.2013.03.003.'''

now = tm.ctime(tm.time())
user = getuser()
formatter = logging.Formatter(' - '.join(
    ["%(asctime)s", "%(name)s", "%(levelname)s", "%(message)s"]))

ch = logging.StreamHandler(sys.stdout)
ch.setFormatter(formatter)
logger = logging.getLogger("metsim")

attrs = {'pet': {'units': 'mm timestep-1',
                 'long_name': 'potential evaporation',
                 'standard_name': 'water_potential_evaporation_flux'},
         'prec': {'units': 'mm timestep-1', 'long_name': 'precipitation',
                  'standard_name': 'precipitation_flux'},
         'shortwave': {'units': 'W m-2', 'long_name': 'shortwave radiation',
                       'standard_name': 'surface_downwelling_shortwave_flux'},
         'longwave': {'units': 'W m-2', 'long_name': 'longwave radiation',
                      'standard_name': 'surface_downwelling_longwave_flux'},
         't_max': {'units': 'C', 'long_name': 'maximum daily air temperature',
                   'standard_name': 'daily_maximum_air_temperature'},
         't_min': {'units': 'C', 'long_name': 'minimum daily air temperature',
                   'standard_name': 'daily_minimum_air_temperature'},
         'temp': {'units': 'C', 'long_name': 'air temperature',
                  'standard_name': 'air_temperature'},
         'vapor_pressure': {'units': 'kPa', 'long_name': 'vapor pressure',
                            'standard_name': 'vapor_pressure'},
         'air_pressure': {'units': 'kPa', 'long_name': 'air pressure',
                          'standard_name': 'air_pressure'},
         'tskc': {'units': 'fraction', 'long_name': 'cloud fraction',
                  'standard_name': 'cloud_fraction'},
         'rel_humid': {'units': '%', 'long_name': 'relative humidity',
                       'standard_name': 'relative_humidity'},
         'spec_humid': {'units': '', 'long_name': 'specific humidity',
                        'standard_name': 'specific_humidity'},
         '_global': {'conventions': '1.6', 'title': 'Output from MetSim',
                     'institution': 'University of Washington',
                     'source': 'metsim.py',
                     'history': 'Created: {0} by {1}'.format(now, user),
                     'references': references,
                     'comment': 'no comment at this time'}}

attrs = {k: OrderedDict(v) for k, v in attrs.items()}


class MetSim(object):
    """
    MetSim handles the distribution of jobs that write to a common file
    by launching muliple processes and queueing up their writeback so that
    work can be done while IO is happening.
    """

    # Class variables
    met_data = None
    methods = {'mtclim': mtclim}
    params = {
        "method": '',
        "domain": '',
        "state": '',
        "out_dir": '',
        "out_state": '',
        "out_prefix": 'forcing',
        "start": '',
        "stop": '',
        "time_step": -1,
        "calendar": 'standard',
        "out_fmt": '',
        "out_precision": 'f4',
        "verbose": 0,
        "sw_prec_thresh": 0.0,
        "time_grouper": None,
        "utc_offset": False,
        "lw_cloud": 'cloud_deardorff',
        "lw_type": 'prata',
        "tdew_tol": 1e-6,
        "tmax_daylength_fraction": 0.67,
        "rain_scalar": 0.75,
        "tday_coef": 0.45,
        "lapse_rate": 0.0065,
        "iter_dims": ['lat', 'lon'],
        "out_vars": ['temp', 'prec', 'shortwave', 'longwave',
                     'vapor_pressure', 'rel_humid']
    }

    def __init__(self, params: dict):
        """
        Constructor
        """
        # Record parameters
        self.params.update(params)

        logger.setLevel(self.params['verbose'])
        ch.setLevel(self.params['verbose'])
        logger.addHandler(ch)
        self._validate_setup()
        logger.info("read_domain")
        self.domain = io.read_domain(self.params)
        self._normalize_times()
        logger.info("read_met_data")
        self.met_data = io.read_met_data(self.params, self.domain)
        self._validate_force_times(force_times=self.met_data['time'])
        logger.info("read_state")
        self.state = io.read_state(self.params, self.domain)
        self.met_data['elev'] = self.domain['elev']
        self.met_data['lat'] = self.domain['lat']
        logger.info("_aggregate_state")
        self._aggregate_state()
        logger.info("load_inputs")
        self.load_inputs()

    def _normalize_times(self):
        # handle when start/stop times are not specified
        for p in ['start', 'stop']:
            # TODO: make sure these options get documented
            if self.params[p] in [None, 'forcing', '']:
                self.params[p] = None
            elif isinstance(self.params[p], str):
                if ':' in self.params[p]:
                    # start from config file
                    date, hour = self.params[p].split(':')
                    year, month, day = date.split('/')
                    self.params[p] = pd.datetime(int(year), int(month),
                                                 int(day), int(hour))
                elif '/' in self.params[p]:
                    # end from config file
                    year, month, day = self.params[p].split('/')
                    self.params[p] = pd.datetime(int(year), int(month),
                                                 int(day))
                else:
                    self.params[p] = pd.to_datetime(self.params[p])
            else:
                self.params[p] = pd.to_datetime(self.params[p])

        logger.info('start {}'.format(self.params['start']))
        logger.info('stop {}'.format(self.params['stop']))

    def _validate_force_times(self, force_times):

        for p, i in [('start', 0), ('stop', -1)]:
            # infer times from force_times
            if self.params[p] is None:
                self.params[p] = pd.Timestamp(
                    force_times.values[i]).to_pydatetime()

        # update calendar from input data (fall back to params version)
        self.params['calendar'] = self.met_data['time'].encoding.get(
            'calendar', self.params['calendar'])

        assert self.params['start'] >= pd.Timestamp(
            force_times.values[0]).to_pydatetime()
        assert self.params['stop'] <= pd.Timestamp(
            force_times.values[-1]).to_pydatetime()

        self.params['state_start'] = (self.params['start'] -
                                      pd.Timedelta("90 days"))
        self.params['state_stop'] = (self.params['start'] -
                                     pd.Timedelta("1 days"))
        logger.info('start {}'.format(self.params['start']))
        logger.info('stop {}'.format(self.params['stop']))

        logger.info('force start {}'.format(pd.Timestamp(
            force_times.values[0]).to_pydatetime()))
        logger.info('force stop {}'.format(pd.Timestamp(
            force_times.values[-1]).to_pydatetime()))

        logger.info('state start {}'.format(self.params['state_start']))
        logger.info('state stop {}'.format(self.params['state_stop']))

        logger.info('calendar {}'.format(self.params['calendar']))
        if self.params['utc_offset']:
            attrs['time'] = {'long_name': 'UTC time',
                             'standard_name': 'utc_time'}
        else:
            attrs['time'] = {'long_name': 'local time at grid location',
                             'standard_name': 'local_time'}

    def load_inputs(self, close=True):
        self.domain = self.domain.load()
        self.met_data = self.met_data.load()
        self.state = self.state.load()
        if close:
            self.domain.close()
            self.met_data.close()
            self.state.close()

    def _get_time_dim_and_time_dim(self):

        index = self.met_data.indexes['time']

        groups = OrderedDict()
        if self.params['time_grouper'] is not None:
            time_dim = pd.Series(index=index)
            grouper = pd.TimeGrouper(self.params['time_grouper'])
            for key, vals in time_dim.groupby(grouper):
                groups[key] = vals.index
        else:
            groups['total'] = index

        return index, groups

    def launch(self):
        """Farm out the jobs to separate processes"""
        # Do the forcing generation and disaggregation if required
        self._validate_setup()
        nprocs = self.params['nprocs']
        iter_list = [self.met_data[dim].values
                     for dim in self.params['iter_dims']]
        self.disagg = int(self.params['time_step']) < cnst.MIN_PER_DAY
        self.method = MetSim.methods[self.params['method']]

        time_dim, groups = self._get_time_dim_and_time_dim()

        for label, times in groups.items():
            self.pool = Pool(processes=nprocs)
            self.site_generator = itertools.product(*iter_list)
            logger.info("Beginning {}".format(label))
            status = []

            # Add in some end point data for continuity in chunking
            times_ext = times.union([times[0] - pd.Timedelta("1 days"),
                                     times[-1] + pd.Timedelta("1 days")]
                                    ).intersection(time_dim)
            data = self.met_data.sel(time=times_ext)
            self.setup_output(self.met_data.sel(time=times))
            for site in self.site_generator:
                locs = {k: v for k, v in zip(self.params['iter_dims'], site)}
                # Don't run masked cells
                if not self.domain['mask'].sel(**locs).values > 0:
                    continue
                stat = self.pool.apply_async(
                    wrap_run,
                    args=(self.method.run, locs, self.params,
                          data.sel(**locs),
                          self.state.sel(**locs),
                          self.disagg, times, label),
                    callback=self._unpack_results)
                status.append(stat)
            self.pool.close()
            # Check that everything worked
            [stat.get() for stat in status]
            self.pool.join()
            self.write(label)

    def run(self):
        """
        Kicks off the disaggregation and queues up data for IO
        """
        self._validate_setup()
        iter_list = [self.met_data[dim].values
                     for dim in self.params['iter_dims']]
        self.disagg = int(self.params['time_step']) < cnst.MIN_PER_DAY
        self.method = MetSim.methods[self.params['method']]

        time_dim, groups = self._get_time_dim_and_time_dim()

        for label, times in groups.items():
            logger.info("Beginning {}".format(label))
            self.site_generator = itertools.product(*iter_list)
            # Add in some end point data for continuity
            times_ext = times.union([times[0] - pd.Timedelta("1 days"),
                                     times[-1] + pd.Timedelta("1 days")]
                                    ).intersection(time_dim)
            data = self.met_data.sel(time=times_ext)
            self.setup_output(self.met_data.sel(time=times))

            for site in self.site_generator:
                locs = {k: v for k, v in zip(self.params['iter_dims'], site)}
                # Don't run masked cells
                if not self.domain['mask'].sel(**locs).values > 0:
                    continue
                wrap_results = wrap_run(
                    self.method.run, locs, self.params, data.sel(**locs),
                    self.state.sel(**locs), self.disagg, times, label)

                # Cut the returned data down to the correct time index
                self._unpack_results(wrap_results)

            self.write(label)

    def _unpack_results(self, result: tuple):
        """Put results into the master dataset"""
        if len(result) == 3:
            locs, df, state = result
            self._unpack_state(state, locs)
        else:
            locs, df = result
        for varname in self.params['out_vars']:
            try:
                self.output[varname].loc[locs] = df[varname]
            except ValueError as e:
                logger.error(e)
                logger.error("This error is probably indicitive of a mismatch "
                             "between the domain and input data. Check that "
                             "all of your cells inside of the mask have both "
                             "elevation in the domain as well as all of the "
                             "required input forcings.")
                raise

    def _unpack_state(self, result: pd.DataFrame, locs: dict):
        """Put restart values in the state dataset"""
        # We concatenate with the old state values in case we don't
        # have 90 new days to use
        tmin = np.concatenate((self.state['t_min'].sel(**locs).values[:],
                               result['t_min'].values))
        tmax = np.concatenate((self.state['t_max'].sel(**locs).values[:],
                               result['t_max'].values))
        prec = np.concatenate((self.state['prec'].sel(**locs).values[:],
                               result['prec'].values))
        self.state['t_min'].sel(**locs).values[:] = tmin[-90:]
        self.state['t_max'].sel(**locs).values[:] = tmax[-90:]
        self.state['prec'].sel(**locs).values[:] = prec[-90:]
        state_start = result.index[-1] - pd.Timedelta('89 days')
        self.state['time'].values = date_range(
            state_start, result.index[-1], calendar=self.params['calendar'])

    def setup_output(self, prototype: xr.Dataset=None):
        if not prototype.variables:
            prototype = self.met_data
        self.disagg = int(self.params['time_step']) < cnst.MIN_PER_DAY
        # Number of timesteps
        if self.disagg:
            delta = pd.Timedelta('1 days') - pd.Timedelta(
                '{} minutes'.format(self.params['time_step']))
        else:
            delta = pd.Timedelta('0 days')

        start = pd.Timestamp(prototype['time'].values[0]).to_pydatetime()
        stop = pd.Timestamp(prototype['time'].values[-1]).to_pydatetime()
        times = date_range(start, stop + delta,
                           freq="{}T".format(self.params['time_step']),
                           calendar=self.params['calendar'])
        n_ts = len(times)

        shape = (n_ts, ) + self.domain['mask'].shape
        dims = ('time', ) + self.domain['mask'].dims
        coords = {'time': times, **self.domain['mask'].coords}
        self.output = xr.Dataset(coords=coords)
        self.output['time'].encoding['calendar'] = self.params['calendar']
        for p in ['elev', 'lat', 'lon']:
            if p in self.params:
                self.params.pop(p)
        for k, v in self.params.items():
            # Need to convert some parameters to strings
            if k in ['start', 'stop', 'time_grouper', 'utc_offset']:
                v = str(v)
            elif k in ['state_start', 'state_stop']:
                # skip
                continue
            # Don't include complex types
            if isinstance(v, dict):
                v = json.dumps(v)
            elif not isinstance(v, str) and isinstance(v, Iterable):
                v = ', '.join(v)
            attrs['_global'][k] = v
        self.output.attrs = attrs['_global']

        dtype = self.params['out_precision']

        for varname in self.params['out_vars']:
            self.output[varname] = xr.DataArray(
                data=np.full(shape, np.nan, dtype=dtype),
                coords=coords, dims=dims,
                name=varname, attrs=attrs.get(varname, {}))
        self.output['time'].attrs.update(attrs['time'])

    def _aggregate_state(self):
        """Aggregate data out of the state file and load it into `met_data`"""
        # Precipitation record

        assert self.state.dims['time'] == 90, self.state['time']

        record_dates = date_range(self.params['state_start'],
                                  self.params['state_stop'],
                                  calendar=self.params['calendar'])
        trailing = self.state['prec']
        trailing['time'] = record_dates
        total_precip = xr.concat([trailing, self.met_data['prec']],
                                 dim='time').load()
        total_precip = (cnst.DAYS_PER_YEAR * total_precip.rolling(
            time=90).mean().sel(time=slice(self.params['start'],
                                           self.params['stop'])))

        self.met_data['seasonal_prec'] = total_precip

        # Smoothed daily temperature range
        trailing = self.state['t_max'] - self.state['t_min']

        trailing['time'] = record_dates
        dtr = self.met_data['t_max'] - self.met_data['t_min']
        if (dtr < 0).any():
            raise ValueError("Daily maximum temperature lower"
                             " than daily minimum temperature!")
        sm_dtr = xr.concat([trailing, dtr], dim='time').load()
        sm_dtr = sm_dtr.rolling(time=30).mean().drop(record_dates, dim='time')
        self.met_data['dtr'] = dtr
        self.met_data['smoothed_dtr'] = sm_dtr

    def _validate_setup(self):
        """Updates the global parameters dictionary"""
        errs = [""]

        # Make sure there's some input
        if not len(self.params.get('forcing', [])):
            errs.append("Requires input forcings to be specified")

        # Parameters that can't be empty strings or None
        non_empty = ['method', 'out_dir', 'out_state', 'start',
                     'stop', 'time_step', 'out_fmt',
                     'forcing_fmt', 'domain_fmt', 'state_fmt']
        for each in non_empty:
            if self.params.get(each, None) is None or self.params[each] == '':
                errs.append("Cannot have empty value for {}".format(each))

        # Make sure time step divides evenly into a day
        if (cnst.MIN_PER_DAY % int(self.params.get('time_step', -1)) or
                (int(self.params['time_step']) > (6 * cnst.MIN_PER_HOUR) and
                 int(self.params['time_step']) != cnst.MIN_PER_DAY)):
            errs.append("Time step must be evenly divisible into 1440 "
                        "minutes (24 hours) and less than 360 minutes "
                        "(6 hours). Got {}.".format(self.params['time_step']))

        # Check for required input variable specification
        if self.met_data is not None:
            required_in = ['t_min', 't_max', 'prec']
            for each in required_in:
                if each not in self.met_data.variables:
                    errs.append("Input requires {}".format(each))

        # Make sure that we are going to write out some data
        if not len(self.params.get('out_vars', [])):
            errs.append("Output variable list must not be empty")

        # Check output variables are valid
        daily_out_vars = ['t_min', 't_max', 'prec', 'vapor_pressure',
                          'shortwave', 'tskc', 'pet', 'wind']
        out_var_check = ['temp', 'prec', 'shortwave', 'vapor_pressure',
                         'air_pressure', 'rel_humid', 'spec_humid',
                         'longwave', 'tskc', 'wind']
        if int(self.params.get('time_step', -1)) == 1440:
            out_var_check = daily_out_vars
        for var in self.params.get('out_vars', []):
            if var not in out_var_check:
                errs.append('Cannot output variable {} at timestep {}'.format(
                    var, self.params['time_step']))

        # Check that the parameters specified are available
        opts = {'out_precision': ['f4', 'f8'],
                'lw_cloud': ['default', 'cloud_deardorff'],
                'lw_type': ['default', 'tva', 'anderson',
                            'brutsaert', 'satterlund',
                            'idso', 'prata']}
        for k, v in opts.items():
            if not self.params.get(k, None) in v:
                errs.append("Invalid option given for {}".format(k))

        # If any errors, raise and give a summary
        if len(errs) > 1:
            raise Exception("\n  ".join(errs))

    def write(self, suffix=""):
        """
        Dispatch to the right function based on the configuration given
        """
        dispatch = {
            'netcdf': self.write_netcdf,
            'ascii': self.write_ascii,
            'data': self.write_data
        }
        dispatch[self.params.get('out_fmt', 'netcdf').lower()](suffix)

    def get_nc_output_suffix(self):
        s, e = self.output.indexes['time'][[0, -1]]
        template = '{:04d}{:02d}{:02d}-{:04d}{:02d}{:02d}'
        return template.format(s.year, s.month, s.day,
                               e.year, e.month, e.day,)

    def write_netcdf(self, suffix: str):
        """Write out as NetCDF to the output file"""
        logger.info("Writing netcdf...")
        for dirname in [self.params['out_dir'],
                        os.path.dirname(self.params['out_state'])]:
            os.makedirs(dirname, exist_ok=True)

        # all state variables are written as doubles
        state_encoding = {}
        for v in self.state.variables:
            state_encoding[v] = {'dtype': 'f8'}
        state_encoding['time']['calendar'] = self.params['calendar']
        # write state file
        self.state.to_netcdf(self.params['out_state'], encoding=state_encoding)

        # write output file
        suffix = self.get_nc_output_suffix()
        fname = '{}_{}.nc'.format(self.params['out_prefix'], suffix)
        output_filename = os.path.join(self.params['out_dir'], fname)
        logger.info(output_filename)
        out_encoding = {'time': {'dtype': 'f8',
                                 'calendar': self.params['calendar']}}
        dtype = self.params['out_precision']
        for v in self.output.data_vars:
            out_encoding[v] = {'dtype': dtype,
                               '_FillValue': cnst.FILL_VALUES[dtype]}
        self.output.to_netcdf(output_filename,
                              unlimited_dims=['time'],
                              encoding=out_encoding)

    def write_ascii(self, suffix):
        """Write out as ASCII to the output file"""
        logger.info("Writing ascii...")
        for dirname in [self.params['out_dir'],
                        os.path.dirname(self.params['out_state'])]:
            os.makedirs(dirname, exist_ok=True)
        # all state variables are written as doubles
        state_encoding = {'time': {'dtype': 'f8'}}
        for v in self.state.variables:
            state_encoding[v] = {'dtype': 'f8'}
        # write state file
        self.state.to_netcdf(self.params['out_state'], encoding=state_encoding)
        # Need to create new generator to loop over
        iter_list = [self.met_data[dim].values
                     for dim in self.params['iter_dims']]
        site_generator = itertools.product(*iter_list)

        for site in site_generator:
            locs = {k: v for k, v in zip(self.params['iter_dims'], site)}
            if not self.domain['mask'].sel(**locs).values > 0:
                continue
            fname = ("{}_" * (len(iter_list) + 1) + "{}.csv").format(
                self.params['out_prefix'], *site, suffix)
            fpath = os.path.join(self.params['out_dir'], fname)
            self.output.sel(**locs)[self.params[
                'out_vars']].to_dataframe().to_csv(fpath)

    def write_data(self, suffix):
        pass


def wrap_run(func: callable, loc: dict, params: dict,
             ds: xr.Dataset, state: xr.Dataset, disagg: bool,
             out_times: pd.DatetimeIndex, group: str):
    """
    Iterate over a chunk of the domain. This is wrapped
    so we can return a tuple of locs and df.

    Parameters
    ----------
    func: callable
        The function to call to do the work
    loc: dict
        Some subset of the domain to do work on
    params: dict
        Parameters from a MetSim object
    ds: xr.Dataset
        Input forcings and domain
    state: xr.Dataset
        State variables at the point of interest
    disagg: bool
        Whether or not we should run a disagg routine
    out_times: pd.DatetimeIndex
        Times to return (should be trimmed 1 day at
        each end from the given index)
    group: str
        The year being run. This is used to add on
        extra times to make output smooth at endpoints
        if the run is chunked in time.

    Returns
    -------
    results
        A list of tuples arranged as
        (location, hourly_output, daily_output)
    """
    logger.info("Processing {}".format(loc))
    lat = ds['lat'].values
    lon = ds['lon'].values
    elev = ds['elev'].values
    params['elev'] = elev
    params['lat'] = lat
    params['lon'] = lon
    df = ds.to_dataframe()
    # solar_geom returns a tuple due to restrictions of numba
    # for clarity we convert it to a dictionary here
    sg = solar_geom(elev, lat, params['lapse_rate'])
    sg = {'tiny_rad_fract': sg[0], 'daylength': sg[1],
          'potrad': sg[2], 'tt_max0': sg[3]}
    yday = df.index.dayofyear.values - 1
    df['daylength'] = sg['daylength'][yday]
    df['potrad'] = sg['potrad'][yday]
    df['tt_max'] = sg['tt_max0'][yday]

    # Generate the daily values - these are saved
    # so that we can use a subset of them to write
    # out the state file later
    df_base = func(df, params)

    if disagg:
        # Get some values for padding the time list,
        # so that when interpolating in the disaggregation
        # functions we can match endpoints with adjoining
        # chunks - if no data is available, just repeat some
        # default values (this case is used at the very
        # beginning and end of the record)
        try:
            prevday = out_times[0] - pd.Timedelta('1 days')
            t_begin = [ds['t_min'].sel(time=prevday),
                       ds['t_max'].sel(time=prevday)]
        except (KeyError, ValueError):
            t_begin = [state['t_min'].values[-1],
                       state['t_max'].values[-1]]
        try:
            nextday = out_times[-1] + pd.Timedelta('1 days')
            t_end = [ds['t_min'].sel(time=nextday),
                     ds['t_max'].sel(time=nextday)]
        except (KeyError, ValueError):
            # None so that we don't extend the record
            t_end = None

        # Disaggregate to subdaily values
        df_complete = disaggregate(df, params, sg, t_begin, t_end)
        # Calculate the times that we want to get out by chopping
        # off the endpoints that were added on previously
        start = out_times[0]
        stop = (out_times[-1] + pd.Timedelta('1 days') -
                pd.Timedelta("{} minutes".format(params['time_step'])))
        new_times = date_range(
            start, stop, freq='{}T'.format(params['time_step']),
            calendar=params['calendar'])
    else:
        # convert srad to daily average flux from daytime flux
        df_base['shortwave'] *= df_base['daylength'] / cnst.SEC_PER_DAY
        # If we're outputting daily values, we dont' need to
        # change the output dates - see inside of `if` condition
        # above for more explanation
        new_times = out_times
        df_complete = df_base

    # Cut the returned data down to the correct time index
    df_complete = df_complete.loc[new_times[0]:new_times[-1]]
    return (loc, df_complete, df_base)
